---
-- Source a complex-valued signal from a PlutoSDR One. This source requires the
-- libPlutoSDR library.
--
-- @category Sources
-- @block PlutoSDRSource
-- @tparam number frequency Tuning frequency in Hz
-- @tparam number rate Sample rate in Hz
-- @tparam[opt={}] table options Additional options, specifying:
--      * `lna_gain` (int, default 8 dB, range 0 to 40 dB, 8 dB step)
--      * `vga_gain` (int, default 40 dB, range 0 to 62 dB, 2 dB step)
--      * `bandwidth` (number in Hz, default round down from sample rate)
--      * `rf_amplifier_enable` (bool, default false)
--      * `antenna_power_enable` (bool, default false)
--
-- @signature > out:ComplexFloat32
--
-- @usage
-- -- Source samples from 135 MHz sampled at 10 MHz
-- local src = radio.PlutoSDRSource(135e6, 10e6)
--
-- -- Source samples from 135 MHz sampled at 10 MHz, with 2.5 MHz bandwidth
-- local src = radio.PlutoSDRSource(135e6, 10e6, {bandwidth = 2.5e6})
--
-- -- Source samples from 91.1 MHz sampled at 8 MHz, with custom gain settings
-- local src = radio.PlutoSDRSource(91.1e6, 8e6, {lna_gain = 16, vga_gain = 22})
--
-- -- Source samples from 144.390 MHz sampled at 8 MHz, with antenna power enabled
-- local src = radio.PlutoSDRSource(144.390e6, 8e6, {antenna_power_enable = true})

local ffi = require('ffi')

local block = require('radio.core.block')
local platform = require('radio.core.platform')
local debug = require('radio.core.debug')
local types = require('radio.types')
local async = require('radio.core.async')
local pipe = require('radio.core.pipe')

local PlutoSDRSource = block.factory("PlutoSDRSource")

function PlutoSDRSource:instantiate(frequency, rate, options)
    self.frequency = assert(frequency, "Missing argument #1 (frequency)")
    self.rate = assert(rate, "Missing argument #2 (rate)")

    self.options = options or {}
    self.lna_gain = self.options.lna_gain or 8
    self.vga_gain = self.options.vga_gain or 40
    self.bandwidth = self.options.bandwidth
    self.rf_amplifier_enable = self.options.rf_amplifier_enable or false
    self.antenna_power_enable = self.options.antenna_power_enable or false

    self:add_type_signature({}, {block.Output("out", types.ComplexFloat32)})
end

function PlutoSDRSource:get_rate()
    return self.rate
end

if not package.loaded['radio.blocks.sinks.plutosdr'] then
    ffi.cdef[[
        enum PlutoSDR_error { PlutoSDR_SUCCESS = 0, };

        typedef struct PlutoSDR_device PlutoSDR_device;

        typedef struct {
            PlutoSDR_device* device;
            char     field2[16];
            uint8_t* buffer;
            int buffer_length;
            int valid_length;
            void* rx_ctx;
            void* tx_ctx;
        } PlutoSDR_transfer;

        typedef struct { uint8_t* version; uint8_t* major; uint8_t* minor; } libv;
 		
        typedef int (*PlutoSDR_sample_block_cb_fn)(PlutoSDR_transfer* transfer);
        const char* PlutoSDR_error_name(enum PlutoSDR_error errcode);
        libv iio_library_get_version(void);
        int PlutoSDR_init(void);
        int PlutoSDR_exit(void);
        int PlutoSDR_create_default_context(void);
        int PlutoSDR_open(PlutoSDR_device** device);
        int PlutoSDR_close(PlutoSDR_device* device);
		int iio_create_default_context(void);
        int PlutoSDR_board_id_read(PlutoSDR_device* device, uint8_t* value);
        const char* PlutoSDR_board_id_name(int board_id);
        int PlutoSDR_version_string_read(PlutoSDR_device* device, char* version, uint8_t length);
		int iio_has_backend(const char *  backend);
		int iio_get_backends_count(void);
	    int iio_context_get_devices_count( 	const struct iio_context *  	ctx	) ;	
        int PlutoSDR_start_rx(PlutoSDR_device* device, PlutoSDR_sample_block_cb_fn callback, void* rx_ctx);
        int PlutoSDR_stop_rx(PlutoSDR_device* device);
        int PlutoSDR_start_tx(PlutoSDR_device* device, PlutoSDR_sample_block_cb_fn callback, void* tx_ctx);
        int PlutoSDR_stop_tx(PlutoSDR_device* device);
        int PlutoSDR_is_streaming(PlutoSDR_device* device);
		int iio_context_find_device(PlutoSDR_device* device);
        int PlutoSDR_set_freq(PlutoSDR_device* device, const uint64_t freq_hz);
        int PlutoSDR_iio_set_sample_rate(PlutoSDR_device* device, const double freq_hz);
        int PlutoSDR_iio_device_find_channel(PlutoSDR_device* device,  	const char *  	name,  	bool  	output );
        int PlutoSDR_set_sample_rate_manual(PlutoSDR_device* device, const uint32_t freq_hz, const uint32_t divider);
        int PlutoSDR_set_baseband_filter_bandwidth(PlutoSDR_device* device, const uint32_t bandwidth_hz);
        int PlutoSDR_set_lna_gain(PlutoSDR_device* device, uint32_t value);
        int PlutoSDR_set_vga_gain(PlutoSDR_device* device, uint32_t value);
        int PlutoSDR_set_txvga_gain(PlutoSDR_device* device, uint32_t value);
        int PlutoSDR_set_amp_enable(PlutoSDR_device* device, const uint8_t value);
        int PlutoSDR_set_antenna_enable(PlutoSDR_device* device, const uint8_t value);
        char* iio_get_backend(char ctx);
        int iio_create_default_context(void);
        char* iio_device_find_channel(char * name);
		uint32_t PlutoSDR_compute_baseband_filter_bw_round_down_lt(const uint32_t bandwidth_hz);
        uint32_t PlutoSDR_compute_baseband_filter_bw(const uint32_t bandwidth_hz);
        int iio_create_context_from_uri(const char *  c_backend);
        char*  iio_context_get_version(const struct iio_context * ctx, unsigned int * major, unsigned int * minor, char git_tag[8] );
      ]]
end
local libPlutoSDR_available, libPlutoSDR = pcall(ffi.load, "iio")

function PlutoSDRSource:initialize()
    -- Check library is available
    if not libPlutoSDR_available then
        error("PlutoSDRSource: libPlutoSDR not found. Is libPlutoSDR installed?")
    end
end

function PlutoSDRSource:initialize_PlutoSDR()
    self.dev = ffi.new("struct PlutoSDR_device *[1]")

 --   local ret

    -- Initialize library
    -- ret = libPlutoSDR.create_context_from_uri("ip:192.168.2.1");
   --     ret = libPlutoSDR.PlutoSDR_iio_get_backends_count();
  --  if ret ~= 0 then
   --     error("PlutoSDR_init(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
   -- end

    -- Open device

 --    ret = libPlutoSDR.iio_library_get_version()
--   ret = libPlutoSDR.iio_context_find_device("local:", "ad9361-phy");
-- debug.printf(ret);
  --  ret = libPlutoSDR.PlutoSDR_open(self.dev)
  -- error("PlutoSDR_open(): " .. ffi.string(libPlutoSDR.iio_library_get_version(ret)))
 --   if ret ~= 0 then
   --     error("PlutoSDR_open(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
  --  end

    -- Dump version info
    if debug.enabled then
        -- Look up firmware version
        local firmware_version = ffi.new("char[128]")
        ret = libPlutoSDR.PlutoSDR_version_string_read(self.dev[0], firmware_version, 128)
        if ret ~= 0 then
            error("PlutoSDR_version_string_read(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
        end
        firmware_version = ffi.string(firmware_version)

        -- Look up board ID
        local board_id = ffi.new("uint8_t[1]")
        ret = libPlutoSDR.PlutoSDR_board_id_read(self.dev[0], board_id)
        if ret ~= 0 then
            error("PlutoSDR_board_id_read(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
        end
        board_id = ffi.string(libPlutoSDR.PlutoSDR_board_id_name(board_id[0]))

        debug.printf("[PlutoSDRSource] Firmware version:  %s\n", firmware_version)
        debug.printf("[PlutoSDRSource] Board ID:          %s\n", board_id)
    end
    local backend = "ip:pluto2.local"
    local c_backend = ffi.new("char[128]", #backend)
    bb = libPlutoSDR.iio_has_backend("serial")
    tt = libPlutoSDR.iio_get_backends_count()
    io.stderr:write(backend)
    backend = libPlutoSDR.iio_get_backend(6)
    io.stderr:write(string.format("Serial backend: %u, Backend count : %u\n",bb, tt))
    io.stderr:write(string.format("Backend6: %s\n",backend))
    io.stderr:write(string.format("Backend1: %s\n",libPlutoSDR.iio_get_backend(1)))
    io.stderr:write(string.format("Backend2: %s\n",libPlutoSDR.iio_get_backend(2)))
    io.stderr:write(string.format("Backend3: %s\n",libPlutoSDR.iio_get_backend(3)))
    libv = libPlutoSDR.iio_library_get_version()
    -- ffi.new("const char*[?]",  #backend)
    io.stderr:write(string.format("LibIIO version : %s\n",libv))
    libPlutoSDR.iio_library_get_version()
        cta = libPlutoSDR.iio_create_default_context()
  ctx = libPlutoSDR.iio_create_context_from_uri(c_backend)



ffi.cdef[[  
static int dev_is_buffer_capable(const struct iio_device *dev)
{
    unsigned int i;
    for (i = 0; i < libPlutoSDR.iio_device_get_channels_count(dev); i++) {
        struct iio_channel *chn = iio_device_get_channel(dev, i);
        printf(" i %s", i);
        if (iio_channel_is_scan_element(chn))
            return true;
    }
    return false;
}

]]            
            io.stderr:write(string.format("cta %s \n   ctx %s\n", cta, ctx))
   --    context = libPlutoSDR.iio_create_context_from_uri(c_backend)
        io.stderr:write(string.format("OK1"))
 --   io.stderr:write(string.format(context))
--    io.stderr:write("OK\n")
 --    vv = libPlutoSDR.iio_context_get_version(libPlutoSDR.iio_get_backend(1))
  --  io.stderr:write(string.format(vv[1]))
  --  io.stderr:write("Library version: %u.%u (git tag: %s)\n", major, minor, git_tag)
    -- Check sample rate
    if self.rate < 8e6 then
        io.stderr:write(string.format("[PlutoSDRSource] Warning: low sample rate (%u Hz).\n", self.rate))
        io.stderr:write("[PlutoSDRSource] Using a sample rate under 8 MHz is not recommended!\n")
    end

    -- Set sample rate
    ret = libPlutoSDR.PlutoSDR_set_sample_rate(self.dev[0], self.rate)
    if ret ~= 0 then
        error("PlutoSDR_set_sample_rate(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end

    -- Compute baseband filter bandwidth
    local computed_bandwidth
    if self.bandwidth then
        -- Snap supplied bandwidth to closest
        computed_bandwidth = libPlutoSDR.PlutoSDR_compute_baseband_filter_bw(self.bandwidth)
    else
        -- Round down from sample rate
        computed_bandwidth = libPlutoSDR.PlutoSDR_compute_baseband_filter_bw_round_down_lt(self.rate)
    end

    debug.printf("[PlutoSDRSource] Frequency: %u Hz, Sample rate: %u Hz\n", self.frequency, self.rate)
    debug.printf("[PlutoSDRSource] Requested Bandwidth: %u Hz, Actual Bandwidth: %u Hz\n", self.bandwidth or computed_bandwidth, computed_bandwidth)

    -- Set baseband filter bandwidth
    ret = libPlutoSDR.PlutoSDR_set_baseband_filter_bandwidth(self.dev[0], computed_bandwidth)
    if ret ~= 0 then
        error("PlutoSDR_set_baseband_filter_bandwidth(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end

    -- Set LNA gain
    ret = libPlutoSDR.PlutoSDR_set_lna_gain(self.dev[0], self.lna_gain)
    if ret ~= 0 then
        error("PlutoSDR_set_lna_gain(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end

    -- Set VGA gain
    ret = libPlutoSDR.PlutoSDR_set_vga_gain(self.dev[0], self.vga_gain)
    if ret ~= 0 then
        error("PlutoSDR_set_vga_gain(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end

    -- Set RF amplifier enable
    ret = libPlutoSDR.PlutoSDR_set_amp_enable(self.dev[0], self.rf_amplifier_enable)
    if ret ~= 0 then
        error("PlutoSDR_set_amp_enable(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end

    -- Set antenna power enable
    ret = libPlutoSDR.PlutoSDR_set_antenna_enable(self.dev[0], self.antenna_power_enable)
    if ret ~= 0 then
        error("PlutoSDR_set_antenna_enable(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end

    -- Set frequency
    ret = libPlutoSDR.PlutoSDR_set_freq(self.dev[0], self.frequency)
    if ret ~= 0 then
        error("PlutoSDR_set_freq(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end
end

local function read_callback_factory(...)
    local ffi = require('ffi')

    local radio = require('radio')
    local pipe = require('radio.core.pipe')

    -- Convert fds on stack to Pipe objects
    local output_pipes = {}
    for i, fd in ipairs({...}) do
        output_pipes[i] = pipe.Pipe()
        output_pipes[i]:initialize(radio.types.ComplexFloat32, nil, fd)
    end

    -- Create pipe mux for write multiplexing
    local pipe_mux = pipe.PipeMux({}, {output_pipes})

    -- Create output vector
    local out = radio.types.ComplexFloat32.vector()

    local function read_callback(transfer)
        -- Resize output vector
        out:resize(transfer.valid_length/2)

        -- Convert complex s8 in buf to complex floats in output vector
        for i = 0, out.length-1 do
            out.data[i].real = ffi.cast("int8_t *", transfer.buffer)[2*i] * (1/127.5)
            out.data[i].imag = ffi.cast("int8_t *", transfer.buffer)[2*i+1] * (1/127.5)
        end

        -- Write to output pipes
        local eof, eof_pipe = pipe_mux:write({out})
        if eof then
            io.stderr:write("[PlutoSDRSource] Downstream block terminated unexpectedly.\n")
        end

        return 0
    end

    return ffi.cast('int (*)(PlutoSDR_transfer *)', read_callback)
end

function PlutoSDRSource:run()
    -- Initialize the PlutoSDR in our own running process
    self:initialize_PlutoSDR()

    -- Create pipe mux for control socket
    local pipe_mux = pipe.PipeMux({}, {}, self.control_socket)

    -- Start receiving
    local read_callback, read_callback_state = async.callback(read_callback_factory, unpack(self.outputs[1]:filenos()))
    local ret = libPlutoSDR.PlutoSDR_start_rx(self.dev[0], read_callback, nil)
    if ret ~= 0 then
        error("PlutoSDR_start_rx(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end

    -- Wait for shutdown from control socket
    while true do
        -- Read control socket
        local _, _, shutdown = pipe_mux:read()
        if shutdown then
            break
        end
    end

    -- Stop receiving
    ret = libPlutoSDR.PlutoSDR_stop_rx(self.dev[0])
    if ret ~= 0 then
        error("PlutoSDR_stop_rx(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end

    -- Close PlutoSDR
    ret = libPlutoSDR.PlutoSDR_close(self.dev[0])
    if ret ~= 0 then
        error("PlutoSDR_close(): " .. ffi.string(libPlutoSDR.PlutoSDR_error_name(ret)))
    end
end

return PlutoSDRSource
